# -*- coding: utf-8 -*-
"""mercat2_report.py: Module for creating tables, figures, and html reports
"""

import base64
import pkg_resources as pkg
import time
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.express as px
from sklearn.decomposition import PCA
from functools import reduce
import dominate
from dominate.tags import *
from dominate.util import raw

from mercat2 import mercat2_metrics


# Global Stylesheet
STYLESHEET = pkg.resource_stream('mercat2', 'data/style.css').read().decode()
LOGO = pkg.resource_stream('mercat2', 'data/logo.jpg').read()
LOGO = base64.b64encode(LOGO).decode()

# TODO: Option for how to include plotly.js.
# False uses script in <head>, 'cdn' loads from internet. # Can I use both???
PLOTLY_SOURCE = 'cdn'


# k-mer summary
def kmer_summary(df_all_samples: dict):
    '''Creates a plotly bar graph with the k-mer counts of the top 2 k-mers among all samples.

    Parameters:
        df_all_samples (dict[name: Dataframe]): A dictionary with the Dataframes of samples containing the counts of k-mers.

    Returns:
        tuple: A tuple with two plotly figures: a barplot of the top k-mers and a plotly table figure labeling the k-mers.
    '''

    numKmers = 5

    df_list = []
    for name,value in df_all_samples.items():
        df_list.append(value['Count'].rename(name))
    df_merged = pd.concat(df_list, axis=1, keys=[s.name for s in df_list]).fillna(0)
    df_merged['Total'] = df_merged.sum(axis=1)
    df_merged['Mean'] = df_merged[[s.name for s in df_list]].mean(axis='columns')
    df_merged.sort_values(by='Mean', axis='index', ascending=False, inplace=True)
    df = df_merged.iloc[0:numKmers][[s.name for s in df_list]].T.reset_index().rename(columns=dict(index='sample'))
    df = df.melt(id_vars=['sample'], var_name='k-mer', value_name='count').sort_values(by='count', ascending=False)
    df['label'] = pd.Categorical(df['k-mer']).codes + 1
    df['label'] = df['label'].apply(lambda x: f'kmer-{x}')
    
    fig = px.bar(df, x='sample', y='count', text='count', color='sample', facet_row='label', template="plotly_white")
    fig.for_each_annotation(lambda a: a.update(text=a.text.replace('label=', '')))
    fig.update_layout(font=dict(color="Black"))
    
    df = df[['k-mer', 'label']].drop_duplicates()

    # Add table to label kmers
    figTable = go.Figure(
        data=[go.Table(cells=dict(values=[df['label'].tolist(), df['k-mer'].tolist()]))],
        layout=go.Layout(margin=go.layout.Margin(l=0, r=0, b=0, t=0,
        ))
    )
    figTable.update_layout(height=100, font=dict(color="Black"), template='plotly_white')
    return [fig, figTable]


# GC Plot kmers
def GC_plot_kmer(df_all_samples:dict):
    '''Creates a plotly bar graph with the GC content of k-mers.

    Parameters:
        df_all_samples (dict[name: Dataframe]): A dictionary with the Dataframes of samples containing the counts of k-mers.

    Returns:
        plotly fig: A plotly barplot figure of the GC% Content of the top k-mers.
    '''

    #print("\nGC PLOT\n")
    df_list = []
    for name,value in df_all_samples.items():
        df_list.append(value['Count'].rename(name))
    df_merged = pd.concat(df_list, axis=1, keys=[s.name for s in df_list]).fillna(0)
    df_merged['Total'] = df_merged.sum(axis=1)
    df_merged['Mean'] = df_merged[[s.name for s in df_list]].mean(axis='columns')
    df_merged.sort_values(by='Mean', axis='index', ascending=False, inplace=True)
    #print(df_merged)
    df = df_merged.reset_index().rename(columns=dict(index='k-mer'))['k-mer'].iloc[0:2]
    def calc_gc(kmer:str):
        return 100.0 * (kmer.count('G') + kmer.count('C')) / len(kmer)
    df = pd.concat([df, df.apply(calc_gc)], axis=1, keys=['k-mer', 'GC'])
    #print("GC CONTENT\n", df)
    fig = px.bar(df, x='k-mer', y='GC', text='GC', template="plotly_white")
    fig.update_layout(font=dict(color="Black"))
    return fig


# GC Plot samples
def GC_plot_sample(gc_content: dict):
    '''Creates a plotly bar graph of the GC content.

    Parameters:
        gc_content (dict[str: float]): A dictionary with GC content data.

    Returns:
        plotly fig: A plotly barplot figure of the GC% Content from the gc_content dictionary.
    '''

    df = pd.DataFrame.from_dict(data=gc_content, orient='index', columns=['GC Content'])
    fig = px.bar(df, template="plotly_white")
    fig.update_layout(font=dict(color="Black"))
    return fig


# Protein Metrics Plot samples
def plot_sample_metrics(protein_samples: dict):
    '''Creates a plotly bar graph of the protein metrics.

    Parameters:
        protein_list (dict[str: list[str]]): A dictionary with lists of sequence files

    Returns:
        plotly fig: A plotly barplot figure of the protein metrics from the samples dictionary.
    '''

    dfMetrics = pd.DataFrame()
    for basename,files in protein_samples.items():
        prot_count = 0
        prot_PI = 0.0
        prot_MW = 0.0
        prot_Hydro = 0.0
        for file in files:
            with open(file) as reader:
                line = reader.readline()
                while line:
                    line = line.strip()
                    line = line.rstrip('*')
                    if line.startswith('>'):
                        name = line[1:]
                        sequence = ""
                        line = reader.readline()
                        while line:
                            line = line.strip()
                            line = line.rstrip('*')
                            if line.startswith('>'):
                                break
                            sequence += line
                            line = reader.readline()
                        prot_count += 1
                        prot_PI = mercat2_metrics.predict_isoelectric_point_ProMoST(sequence)
                        prot_MW = mercat2_metrics.calculate_MW(sequence)
                        prot_Hydro = mercat2_metrics.calculate_hydro(sequence)
                        continue #already got next line, next item in loop
                    line = reader.readline()
        dfMetrics.at[basename, 'PI Avg'] = prot_PI / prot_count
        dfMetrics.at[basename, 'MW Avg'] = prot_MW / prot_count
        dfMetrics.at[basename, 'Hydro Avg'] = prot_Hydro / prot_count

    figPI = px.bar(dfMetrics, x=dfMetrics.index, y='PI Avg', template="plotly_white",
        labels={'index':'Sample', 'PI Avg':'PI Average'})
    figPI.update_layout(font=dict(color="Black"))
    
    figMW = px.bar(dfMetrics, x=dfMetrics.index, y='MW Avg', template="plotly_white",
        labels={'index':'Sample', 'MW Avg':'MW Average'})
    figMW.update_layout(font=dict(color="Black"))
    
    figHydro = px.bar(dfMetrics, x=dfMetrics.index, y='Hydro Avg', template="plotly_white",
        labels={'index':'Sample', 'Hydro Avg':'Hydropathy Average'})
    figHydro.update_layout(font=dict(color="Black"))
    return ([figPI, figMW, figHydro], dfMetrics)


# PCA
def PCA_plot(dfPCA):
    '''Creates a 3D plotly scatter plot of the PCA of the given Dataframe.

    Parameters:
        dfPCA (Dataframe): A pandas dataframe with samples and kmer counts

    Returns:
        plotly fig: A plotly 3D scatter plot of the PCA of the dataframe.
    '''

    df_merged2 = dfPCA
    result = df_merged2.replace(np.nan, 0)
    pivoted = result.T
    res = pivoted.rename(columns=pivoted.iloc[0])
    res1 = res.drop(res.index[0])
    pca = PCA(n_components=3, svd_solver='randomized')
    X_train = pca.fit_transform(res1)
    labels = {
    str(i): f"PC {i+1} ({var:.1f}%)"
    for i, var in enumerate(pca.explained_variance_ratio_ * 100)
    }
    figPCA = px.scatter_3d(
        X_train, x=0, y=1, z=2, color=res1.index,
        labels=labels,
        template="plotly_white"
    )
    figPCA.update_layout(font=dict(color="Black"),
        margin=dict(l=0, r=0, t=0, b=0),
        )
    
    return figPCA


# Write HTML Report
def write_html(outfile:str, figPlots:dict, tsv_stats:dict):
    '''Creates an html report with the figures and tsv files.

    Parameters:
        outfile (str): A path to the html file to save the report.
        figPlots (dict): A dictionary of plotly figures.
        tsv_stats (dict): A dictionary of paths to TSV files to add to the downloads section.

    Returns:
        None
    '''

    # WRITE TO FORMATED HTML
    with dominate.document(title='K-Mer Report') as doc:
        # Header
        with doc.head:
            meta(charset="utf-8")
            script(type="text/javascript", src="plotly-2.0.0.min.js")
            with style(type="text/css"):
                raw('\n'+STYLESHEET)
        # Body of HTML
        with div(cls="document", id="mercat-summary"):
            # Title bar at the top
            with h1(cls="title"):
                img(src=f"data:image/png;base64,{LOGO}", height="40")
                a("MERCAT2", cls="reference external", href="https://github.com/raw-lab/mercat2")
                raw(" - Summary")
            # Table of contents - left panel
            with div(cls="contents topic", id="contents"):
                with ul(cls="simple"):
                    li(a("Summary", cls="reference internal", href="#summary"))
                    with ul():
                        for key in figPlots.keys():
                            li(a(f"{key}", cls="reference internal", href=f"#{key}"))
                    li(a("Downloads", cls="reference internal", href="#downloads"))
            # Main panel
            with div(h1("Summary"), cls="section", id="summary"):
                for key,figures in figPlots.items():
                    with div(h2(f"{key}"), cls="section", id=f"{key}"):
                        if type(figures) is list:
                            for fig in figures:
                                raw(fig.to_html(full_html=False, include_plotlyjs=PLOTLY_SOURCE))
                        else:
                            raw(figures.to_html(full_html=False, include_plotlyjs=PLOTLY_SOURCE))
            # Downloads at the bottom
            with div(cls="section", id="downloads"):
                h1("Downloads")
                with div(cls="docutils container", id="attachments"):
                    with blockquote():
                        with div(cls="docutils container", id="table-1"):
                            with dl(cls="docutils"):
                                for key,value in tsv_stats.items():
                                    data_URI = f"data:text/tab-separated-values;base64,{value}"
                                    dt(key)
                                    dd(a(key+".tsv", href=data_URI, download=key+".tsv", draggable="true"))
                div(time.strftime("%Y-%m-%d", time.localtime()), cls="docutils container", id="metadata")

    with open(outfile, 'w') as writer:
        writer.write(doc.render())
    return



# -*- coding: utf-8 -*-
"""mercat2_fasta.py: Module with tools for fasta files
"""

import os
import subprocess
import re
import textwrap


## Split Sequence by N
def split_sequenceN(header:str, sequence:str):
    '''Splits a nucleotide sequence string at occurrences of N.
    Helper method for removeN.
    Assumes the first word before the first space is the name of the sequence.

    Parameters:
        header (str): The header of the sequence, without the leading '>'.
        sequence (str): The nucleotide sequence.

    Returns:
        tuple: A tuple with a list of sequences (including modified headers)
               and a corresponding list of contiguous N repeats in each sequence.
    '''

    N_lengths = []
    regex = re.compile(r"(N+)")
    for match in regex.finditer(sequence):
        N_lengths.append(len(match.group(1)))
    sequences = regex.sub('\n', sequence).split('\n')
    header = header.split()
    basename = header[0]
    info = ' '.join(header[1:])
    seqs = []
    for i, seq in enumerate(sequences, 1):
        header = f">{basename}_{i} {info}"
        seqs.append(header)
        seqs += textwrap.wrap(seq, 80)
    
    return (seqs, N_lengths)


## Remove N's
def removeN(fasta:str, outpath:str):
    '''Splits sequences in a scaffold fasta file at N repeats.

    Parameters:
        fasta (str): A path to a fasta file.
        outpath (str): A folder path of where to save the modified file.

    Returns:
        tuple: A tuple with the path to the cleaned file and a dictionary containing basic stats.
    '''

    os.makedirs(outpath, exist_ok=True)    

    outFasta, ext = os.path.splitext(fasta)
    outFasta = os.path.basename(outFasta) + "_clean"+ ext
    outFasta = os.path.join(outpath, outFasta)

    NStats = dict()
    gc_count = 0
    total_length = 0
    with open(fasta) as reader, open(outFasta, 'w') as writer:
        line = reader.readline()
        while line:
            line = line.strip()
            if line.startswith('>'):
                name = line[1:]
                sequence = ""
                line = reader.readline()
                while line:
                    line = line.strip()
                    if line.startswith('>'):
                        break
                    sequence += line
                    line = reader.readline()
                if 'N' in sequence:
                    sequences, stats = split_sequenceN(name, sequence)
                    print('\n'.join(sequences), file=writer)
                    for seq in sequences:
                        gc_count += seq.count('G') + seq.count('C')
                        total_length += len(seq)
                else:
                    print('>', name, sep='', file=writer)
                    print('\n'.join(textwrap.wrap(sequence, 80)), file=writer)
                    gc_count += sequence.count('G') + sequence.count('C')
                    total_length += len(sequence)
                continue #already got next line, next item in loop
            line = reader.readline()
    NStats['GC Content'] = 100.0 * gc_count / total_length

    return (os.path.abspath(outFasta), NStats)


## Process Fastq
def fastq_processing(fq_file:str, outpath:str, f_name:str):
    '''Processes fastq files.
    Uses fastqc to get the quality of the reads.
    Uses fastp to trim the reads.
    Uses Linux sed command to convert the fastq file to fasta format.

    Parameters:
        fq_file (str): The path to a fastq file.
        outpath (str): The path to save the fastqc report, trimmed fastq file, and fasta file.
        f_name (str): The name of the sample.

    Returns:
        str: The absolute path to the final trimmed fasta file.
    '''

    os.makedirs(outpath, exist_ok=True)
    trim_fq = os.path.join(outpath, f_name+"_trim.fastq")
    trim_fna = os.path.join(outpath, f_name+"_trim.fna")
    
    cmd_qc = f"fastqc {fq_file} -o {outpath}"
    cmd_fastp = f"fastp -i {fq_file} -o {trim_fq}"
    cmd_qc2 = f"fastqc {trim_fq} -o {outpath}"
    cmd_format = f"sed -n '1~4s/^@/>/p;2~4p' {trim_fq} > {trim_fna}"
    subprocess.call(cmd_qc, shell=True)
    subprocess.call(cmd_fastp, shell=True)
    subprocess.call(cmd_qc2, shell=True)
    subprocess.call(cmd_format, shell=True)
    return os.path.abspath(trim_fna)


## ORF Call Prodigal
def orf_call(basename:str, file:str, outpath:str):
    '''Finds the ORFs in the nucleotides and converts to an amino acid file.
    Uses prodigal for ORF calling.
    Produces a protein ffa file.
    Produces a gff file.

    Parameters:
        basename (str): The base name of the file for output files.
        file (str): The path to a nucleotide fasta file.
        outpath (str): The path to save the output files.

    Returns:
        tuple: A tuple with the name, and path to the protein faa file.
    '''

    outpath = os.path.abspath(outpath)
    out_pro = os.path.join(outpath, basename+"_pro.faa")
    out_gff = os.path.join(outpath, basename+".gff")
    out_nuc = os.path.join(outpath, basename+"_nuc.fna")
    prod_cmd = f"prodigal -i {file} -o {out_gff} -a {out_pro} -f gff -p meta"
    os.makedirs(outpath, exist_ok=True)
    with open(os.devnull, 'w') as FNULL:
        subprocess.call(prod_cmd, stdout=FNULL, stderr=FNULL, shell=True)
    return (basename, out_pro)



# -*- coding: utf-8 -*-
"""mercat2_kmers.py: Module for calculating k-mer counts
"""

## Calculate k-mer count
#
def calculateKmerCount(seq:str, kmer:int) -> dict:
    '''Calculates the k-mer count in a sequence.
    This is a helper method for find_kmers()

    Parameters:
        seq (str): The sequence to search for k-mers.
        kmer (int): The k-mer size

    Returns:
        dict: A dictionary with the counts of each k-mer found.
    '''

    kmerlist = dict()
    for i in range((len(seq) - kmer) + 1):
        k = seq[i:i+kmer]
        if k not in kmerlist:
            kmerlist[k] = 0
        kmerlist[k] += 1
    return kmerlist


## Find k-mers
def find_kmers(file:str, kmer:int, min_count:int, num_cores:int):
    '''Calculates the k-mer count in a fasta file.

    Parameters:
        file (str): path to a fasta file to scan for k-mers.
        kmer (int): k-mer length.
        min_count (int): minimum count of k-mers found to be considered significant.
        num_cores (int): cpu count to use for parallel processing.

    Returns:
        dict: A dictionary with the counts of each k-mer found.
    '''

    num = 0
    kmerlist = dict()
    with open(file, 'r') as f:
        seq = ""
        for line in f:
            line = line.strip()
            if line.startswith(">"):
                num += 1
                if seq:
                    # count k-mers
                    for i in range((len(seq) - kmer) + 1):
                        k = seq[i:i+kmer]
                        if k not in kmerlist:
                            kmerlist[k] = 0
                        kmerlist[k] += 1
                    seq = ""
            else:
                seq += line.replace("*","")
        # Count last sequence
        for i in range((len(seq) - kmer) + 1):
            k = seq[i:i+kmer]
            if k not in kmerlist:
                kmerlist[k] = 0
            kmerlist[k] += 1
    
    significant_kmers = dict()
    for k,v in kmerlist.items():
        if kmerlist[k] >= min_count:
            significant_kmers[k] = v

    return significant_kmers



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""mercat2-pipeline.py: Python code for Parallel k-mer counting."""

__version__     = "0.1"
__author__      = "Jose L. Figueroa III, Richard A. White III"
__copyright__   = "Copyright 2022"

import sys
import os
import psutil
import shutil
import subprocess
import ray
import pandas as pd
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import timeit

# Mercat libraries
from mercat2 import (mercat2_fasta, mercat2_Chunker, mercat2_kmers, mercat2_report)


# GLOBAL VARIABLES
FILE_EXT_FASTQ = ['.fastq']
FILE_EXT_NUCLEOTIDE = [".fasta", ".fa", ".fna", ".ffn"]
FILE_EXT_PROTEIN = [".faa"]


## Check Command
#
def check_command(cmd:str):
    '''Raises an error and exists if the given command fails'''
    try:
        subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)
    except subprocess.CalledProcessError as e:
        print(f"Mercat Error: {cmd} not found, please setup {cmd} using: 'conda install {cmd}'")
        sys.exit(1)
    return


## Parse Arguments
#
def parseargs():
    '''Returns the parsed command line options.'''
    num_cores = psutil.cpu_count(logical=False)
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-i', type=str, required=False, help='path-to-input-file') #default=nucleotide
    parser.add_argument('-f', type=str, required=False, help='path-to-folder-containing-input-files')
    parser.add_argument('-k', type=int, required = True, help='kmer length')
    parser.add_argument('-n', type=int, default=num_cores, help='no of cores [default = all]')  # no of cores to use
    parser.add_argument('-c', type=int, default=10, help='minimum kmer count [default = 10]')  # minimum kmer count to report
    parser.add_argument('-prod', action='store_true', help='run prodigal on fasta file')
    parser.add_argument('-s', type=int, default=100, required=False, help='Split into x MB files. Default = 100MB')
    parser.add_argument('-o', type=str, default='mercat_results', required=False, help="Output folder, default = 'mercat_results' in current directory")

    # Process arguments
    args = parser.parse_args()

    # Check folder/input file options
    if args.i and args.f:
        parser.error("Can only specify either an input file (-i) or path to folder containing input files (-f) at a time")
    if args.i:
        if not os.path.isfile(args.i):
            parser.error(f"file {args.i} is not valid.\n")
    elif args.f:
        if not os.path.isdir(args.f):
            parser.error(f"folder {args.f} is not valid.\n")
    else:
        parser.error("Please provide either an input file (-i) or an input folder (-f)")

    # check prodigal/protein flags
    if args.prod:
        check_command('prodigal')
    return args


## Chunk Files
#
@ray.remote(num_cpus=1)
def chunk_files(name:str, file:str, chunk_size:int, outpath:str):
    '''Checks if the given file is larger than chunk_size and splits the file as necessary.
    Fasta files are split at sequence headers to keep individual sequences contiguous.

    Parameters:
        name (str): The name of the sample file for proper grouping of the files.
        file (str): Path to a fasta.
        chunk_size (int): The size, in MB, to split the files at.
        outpath (str): Path to a folder to save the chunked files in. Folder should not exist.

    Returns:
        tuple: A tuple with the name, and list of paths to the chunked files.
    '''

    if os.stat(file).st_size >= (chunk_size*1024*1024):
        print("Large input file provided: Splitting it into smaller files...\n")
        os.makedirs(outpath, exist_ok=True)
        all_chunks = mercat2_Chunker.Chunker(file, outpath, str(chunk_size)+"M", ">").files
    else:
        all_chunks = [file]
    return (name, all_chunks)


## Main
#
def mercat_main():
    '''Main method'''

    # Process Arguments
    __args__ = parseargs()
    # Set Flags
    m_kmer = __args__.k
    m_num_cores = __args__.n
    m_inputfile = __args__.i
    m_inputfolder = __args__.f
    m_min_count = __args__.c
    m_flag_prodigal = __args__.prod
    m_chunk_size = __args__.s
    m_outputfolder = __args__.o
    
    if os.path.exists(m_outputfolder) and os.path.isdir(m_outputfolder):
        shutil.rmtree(m_outputfolder)
    os.makedirs(m_outputfolder, exist_ok=True)

    # Initialize Ray
    try:
        ray.init(address='auto') # First try if ray is setup for a cluster
    except:
        ray.init()

    # Load input files
    all_ipfiles = []
    gc_content = dict()
    files_nucleotide = {}
    files_protein = {}
    cleanpath = os.path.join(m_outputfolder, 'clean')
    print("Loading files")
    if m_inputfolder:
        m_inputfolder = os.path.abspath(m_inputfolder)
        for fname in os.listdir(m_inputfolder):
            file = os.path.join(m_inputfolder, fname)
            if not os.path.isdir(file):
                # skip directories
                basename, f_ext = os.path.splitext(fname)
                if f_ext in FILE_EXT_FASTQ:
                    check_command('fastqc')
                    check_command('fastp')
                    file = mercat2_fasta.fastq_processing(file, cleanpath, basename)
                    files_nucleotide[basename] = file
                elif f_ext in FILE_EXT_NUCLEOTIDE:
                    file,stat = mercat2_fasta.removeN(file, cleanpath)
                    files_nucleotide[basename] = file
                    gc_content[basename] = stat['GC Content']
                elif f_ext in FILE_EXT_PROTEIN:
                    files_protein[basename] = file
                all_ipfiles.append(file)
    else:
        basename, f_ext = os.path.splitext(m_inputfile)
        if f_ext in FILE_EXT_FASTQ:
            check_command('fastqc')
            check_command('fastp')
            m_inputfile = mercat2_fasta.fastq_processing(m_inputfile, cleanpath, basename)
            files_nucleotide[basename] = m_inputfile
        elif f_ext in FILE_EXT_NUCLEOTIDE:
            m_inputfile,stat = mercat2_fasta.removeN(m_inputfile, cleanpath)
            files_nucleotide[basename] = m_inputfile
            gc_content[basename] = stat['GC Content']
        elif f_ext in FILE_EXT_PROTEIN:
            files_protein[basename] = m_inputfile
        all_ipfiles.append(m_inputfile)
        all_ipfiles.append(os.path.abspath(m_inputfile))

    
    # ORF Call if -prod flag
    if m_flag_prodigal:
        @ray.remote(num_cpus=1)
        def orf_call(basename, file, prodpath):
            return mercat2_fasta.orf_call(basename, file, prodpath)

        print(f"Running prodigal on {len(files_nucleotide)} files")
        prodpath = os.path.join(m_outputfolder, 'prodigal')
        jobs = []
        for basename,file in files_nucleotide.items():
            jobs += [orf_call.remote(basename, file, prodpath)]
        while jobs:
            ready,jobs = ray.wait(jobs)
            name,amino = ray.get(ready[0])
            files_protein[name] = amino


    # Chunk large files
    print("Checking for large files")
    dir_chunks = os.path.join(m_outputfolder, "chunks")
    # nucleotides
    jobs = []
    for basename,file in files_nucleotide.items():
        jobs += [chunk_files.remote(basename, file, m_chunk_size, os.path.join(dir_chunks, f"{basename}_nucleotide"))]
    while jobs:
        ready,jobs = ray.wait(jobs)
        name,chunks = ray.get(ready[0])
        files_nucleotide[name] = chunks
    # proteins
    jobs = []
    for basename,file in files_protein.items():
        jobs += [chunk_files.remote(basename, file, m_chunk_size, os.path.join(dir_chunks, f"{basename}_protein"))]
    while jobs:
        ready,jobs = ray.wait(jobs)
        name,chunks = ray.get(ready[0])
        files_protein[name] = chunks

    # Begin processing files
    figPlots = dict()
    tsv_stats = dict()
    os.makedirs(os.path.join(m_outputfolder, 'tsv'), exist_ok=True)

    @ray.remote(num_cpus=1)
    def countKmers(file, kmer, min_count, num_cores):
        return mercat2_kmers.find_kmers(file, kmer, min_count, num_cores)

    # Process Nucleotides
    print("Processing Nucleotides")
    df_list = dict()
    for basename,files in files_nucleotide.items():
        np_string = '_nucleotide'

        # Run Mercat
        print(f"Running mercat using {m_num_cores} cores on {basename}{np_string}")
        start_time = timeit.default_timer()
        kmers = dict()
        jobs = []
        for file in files:
            jobs += [countKmers.remote(file, m_kmer, m_min_count, m_num_cores)]
        while(jobs):
            ready,jobs = ray.wait(jobs)
            for k,v in ray.get(ready[0]).items():
                if k in kmers:
                    kmers[k] += v
                else:
                    kmers[k] = v
        print(f"Significant k-mers: {len(kmers)}")
        print(f"Time to compute {m_kmer}-mers: {round(timeit.default_timer() - start_time,2)} secs")
        df_list[basename] = pd.DataFrame(index=kmers.keys(), data=kmers.values(), columns=['Count'])
        outfile = os.path.join(m_outputfolder, 'tsv', f"{basename}{np_string}_summary.tsv")
        dfGC = df_list[basename].reset_index().rename(columns=dict(index='k-mer'))
        dfGC['GC%'] = dfGC['k-mer'].apply(lambda k: round(100.0 * (k.count('G') + k.count('C')) / len(k), 2))
        dfGC.to_csv(outfile, index=False, sep='\t')
    # Stacked Bar Plots (top kmer counts)
    if len(df_list):
        figPlots["Combined Nucleotide kmer Summary"] = mercat2_report.kmer_summary(df_list)
        if len(files_nucleotide) > 3:
            dfPCA = pd.DataFrame()
            for name,df in df_list.items():
                dfTemp = df.rename(columns=dict(Count=name))
                dfPCA = dfPCA.merge(dfTemp, left_index=True, right_index=True, how="outer")
            dfPCA = dfPCA.reset_index().rename(columns=dict(index="k-mer"))
            figPlots["Nucleotide PCA"] = mercat2_report.PCA_plot(dfPCA)
        if m_kmer >= 10:
            figPlots['k-mer GC Summary'] = mercat2_report.GC_plot_kmer(df_list)
    if len(gc_content) > 0:
        figPlots['Sample GC Summary'] = mercat2_report.GC_plot_sample(gc_content)
    

    # Process Proteins
    print("Processing Proteins")
    df_list = dict()
    for basename,files in files_protein.items():
        np_string = '_protein'

        # Run Mercat
        start_time = timeit.default_timer()
        kmers = dict()
        jobs = []
        for file in files:
            jobs += [countKmers.remote(file, m_kmer, m_min_count, m_num_cores)]
        while(jobs):
            ready,jobs = ray.wait(jobs)
            for k,v in ray.get(ready[0]).items():
                if k in kmers:
                    kmers[k] += v
                else:
                    kmers[k] = v
        print("Significant k-mers:", len(kmers))
        print(f"Time to compute {m_kmer}-mers: {round(timeit.default_timer() - start_time,2)} secs")
        df_list[basename] = pd.DataFrame(index=kmers.keys(), data=kmers.values(), columns=['Count'])
        outfile = os.path.join(m_outputfolder, 'tsv', f"{basename}{np_string}_summary.tsv")
        df_list[basename].to_csv(outfile, index=False, sep='\t')
    # Stacked Bar Plots (top kmer counts)
    if len(df_list):
        figPlots["Combined Protein kmer Summary"] = mercat2_report.kmer_summary(df_list)
        if len(files_protein) > 3:
            dfPCA = pd.DataFrame()
            for name,df in df_list.items():
                df = df.rename(columns=dict(Count=name))
                dfPCA = dfPCA.merge(df, left_index=True, right_index=True, how="outer")
            dfPCA = dfPCA.reset_index().rename(columns=dict(index="k-mer"))
            figPlots["Protein PCA"] = mercat2_report.PCA_plot(dfPCA)
    if len(files_protein):
        figPlots["Sample Protein Metrics Summary"],tsv_stats["Protein_Metrics"] = mercat2_report.plot_sample_metrics(files_protein)
    
    
    # Plot Data
    plots_dir = os.path.join(m_outputfolder, "plots")
    os.makedirs(plots_dir, exist_ok=True)

    mercat2_report.write_html(os.path.join(plots_dir, "report.html"), figPlots, tsv_stats)
    return


# If main app
if __name__ == "__main__":
    mercat_main()



import os
import setuptools

#Anaconda dependencies = "fastqc fastp prodigal"
#$ conda create -n mercat -c bioconda python=3.7 fastqc fastp prodigal -y


# recursively load package files
def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


setuptools.setup(
    name = "mercat2",
    version = "0.1",
    author = "Richard White III",
    author_email = "rwhit101@uncc.edu",
    description = "versatile k-mer counter and diversity estimator for database independent property analysis (DIPA) for multi-omic analysis",
    long_description = open("README.md", "r").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/raw-lab/mercat2",
    scripts = ['bin/mercat2-pipeline.py'], # scripts to install to 'bin' path
    packages = ['mercat2'], # list of packages, installed to 'site-packages' folder
    package_dir = dict(mercat2='mercat2'), # dict with 'package'='relative dir'
    package_data = dict(mercat2=package_files('mercat2/data')), # add non-python data to package, relative paths
    license = "MIT License", # metadata
    platforms = ['Unix'], # metadata
    classifiers = [ # This is the new updated way for metadata, but old way seems to still be used in some of the output
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    python_requires = '<=3.10',
    install_requires = [
            'setuptools',
            'ray',
            'configargparse',
            'pandas',
            'numpy',
            'humanize',
            'plotly',
            'psutil',
            'joblib',
            'dominate',
            'scikit-learn',
    ]
)
